import 'dart:typed_data';

class Quaternion {
  double w;
  double x;
  double y;
  double z;

  Quaternion(this.w, this.x, this.y, this.z);
}

class EulurAngle {
  double x;
  double y;
  double z;

  EulurAngle(this.x, this.y, this.z);
}

class Vector3 {
  double x;
  double y;
  double z;

  Vector3(this.x, this.y, this.z);
}

class SamplingData {
  double stretch;
  Quaternion quaternion;
  EulurAngle eulurAngle;
  double deltaTime;
}

class HiflexSampling {
  Map<String, HiflexScalar> _rawData = new Map();
  DateTime _privTime = new DateTime.now();

  HiflexSampling() {
    _rawData.addAll({
      "stretch": new HiflexScalarKalman(),
      "acc.x": new HiflexScalarNormal(),
      "acc.y": new HiflexScalarNormal(),
      "acc.z": new HiflexScalarNormal(),
      "gyr.x": new HiflexScalarNormal(),
      "gyr.y": new HiflexScalarNormal(),
      "gyr.z": new HiflexScalarNormal(),
      "mag.x": new HiflexScalarNormal(),
      "mag.y": new HiflexScalarNormal(),
      "mag.z": new HiflexScalarNormal(),
      "deltaTime": new HiflexScalarKalman(),
    });
  }

  SamplingData get data {
    var sampling = new SamplingData();
    sampling.stretch = _rawData["stretch"].load();
    sampling.deltaTime = _rawData["deltaTime"].load();
    //sampling.quaternion;

    return sampling;
  }

  ///
  SamplingData load(Uint8List rawData) {
    // Stream으로 전달.???
    var bytes = ByteData.sublistView(rawData);

    String printString = "";
    for (int n = 0; n < bytes.lengthInBytes; ++n) {
      printString += "${bytes.getUint8(n).toRadixString(16)}:";
    }
    print("Recv Data[${rawData.length}] $printString");

    // 여기서 Quaternion 계산 후 _rawData에 저장.
    _rawData["acc.x"].store(bytes.getUint16(3, Endian.big).toDouble());
    _rawData["acc.y"].store(bytes.getUint16(5, Endian.big).toDouble());
    _rawData["acc.z"].store(bytes.getUint16(7, Endian.big).toDouble());
    _rawData["gyr.x"].store(bytes.getUint16(9, Endian.big).toDouble());
    _rawData["gyr.y"].store(bytes.getUint16(11, Endian.big).toDouble());
    _rawData["gyr.z"].store(bytes.getUint16(13, Endian.big).toDouble());
    _rawData["mag.x"].store(bytes.getUint16(15, Endian.big).toDouble());
    _rawData["mag.y"].store(bytes.getUint16(17, Endian.big).toDouble());
    _rawData["mag.z"].store(bytes.getUint16(19, Endian.big).toDouble());

    _rawData["stretch"].store(bytes.getUint16(21, Endian.big).toDouble());

    var nowTime = new DateTime.now();
    _rawData["deltaTime"]
        .store(nowTime.difference(_privTime).inMilliseconds.toDouble());
    _privTime = nowTime;

    return this.data;
  }
}

/*
 * 칼만 필터를 적용해야 하는 부분이 있음.
 */
class HiflexScalar {
  double load() => 0.0;
  store(double changeValue) {}
}

abstract class HiflexScalarBase implements HiflexScalar {}

class HiflexScalarNormal extends HiflexScalarBase {
  double _value;

  double load() => _value;
  store(double changeValue) {
    _value = changeValue;
  }
}

class HiflexScalarKalman extends HiflexScalarBase {
  double _errMeasure = 2.0;
  double _errEstimate = 2.0;
  double _q = 0.01;
  double _currentEstimate = 0.0;
  double _lastEstimate = 0.0;
  double _kalmanGain = 0.0;

  double load() => _currentEstimate;
  store(double changeValue) {
    _kalmanGain = _errEstimate / (_errEstimate + _errMeasure);
    _currentEstimate =
        _lastEstimate + _kalmanGain * (changeValue - _lastEstimate);
    _errEstimate = (1.0 - _kalmanGain) * _errEstimate +
        (_lastEstimate - _currentEstimate).abs() * _q;
    _lastEstimate = _currentEstimate;
  }
}
